/*
 * File : dataPacket.C
 *
 * Description : This file contains the method definitions for the
 * 	dataPacket class.
 */

#include <memory.h>
#include <string.h>
#include <iostream.h>
#include "trace.h"
#include "types.h"
#include "dataPacket.h"

DEFINE_TRACE_MODULE(dataPacketModule,"dataPacketModule");

/*
 * Function dataPacket()
 *
 * Description : This is the constructor which takes no arguments.
 *	It allocates buffer space equal to MaxPacketSize, and initializes
 *	the buffer pointers.
 */

dataPacket::dataPacket()
{
	TRACE(dataPacketModule,2,("Constructing data packet\n"));
	_buf = new unsigned char[MaxPacketSize];
	_head = _buf;
	_tail = _buf;
	_count = 0;

	TRACE(dataPacketModule,1,("Head: %d Tail: %d Buf: %d\n",_head,_tail,_buf)); 
}

/*
 * Function dataPacket(const unsigned char *data, int len)
 *
 * Description : This is the constructor which takes arguments.
 *	It allocates buffer space for the data, copies data into
 *	the buffer, and initializes the buffer pointers.
 *	BIG ASSUMPTION : the length of data is less than MaxPacketSize -
 *	less enough so that even after all additional data has been
 *	appended and prepended, the length will still be <= MaxPacketSize.
 */

dataPacket::dataPacket(const unsigned char *data, int len )
{
	TRACE(dataPacketModule,2,("Constructing data packet\n"));
	_buf = new unsigned char[MaxPacketSize];
	InsertNewData(data,len);
	_count = 0;
}

/*
 * Function : ~dataPacket()
 *
 * Description : This is the destructor. It frees allocated memory.
 */

dataPacket::~dataPacket()
{
	TRACE(dataPacketModule,2,("Destroying data packet\n"));
	delete _buf;
}

/*
 * Function : Prepend()
 *
 * Description : This function prepends a given amount of data
 * to the buffer. It copies the additional data into the buffer
 * and rearranges the pointers. 
 * Returns TRUE if successful, FALSE if failure.
 */

Boolean
dataPacket::Prepend(const unsigned char *data, int len)
{
	TRACE(dataPacketModule,1,("In Prepend\n"));
	if ( _head == _tail && _head == _buf ) { 
		// Nothing is in the buffers..
		return InsertNewData(data,len);
	}
	else {
		if ( (_head - _buf) < len ) {
			// Not enough space
			TRACE(dataPacketModule,1,("Data larger than MaxPacketSize\n"));
			return FALSE;
		}
		// Modify pointer and copy data.
		_head -= len;
		memcpy(_head,data,len);
	}
	TRACE(dataPacketModule,1,("Head: %d Tail: %d Buf: %d\n",_head,_tail,_buf)); 
	return TRUE;
}

/*
 * Function : Append()
 *
 * Description : This function appends a given amount of data
 * to the buffer. It copies the additional data into the
 * buffer and rearranges the pointers.
 * Returns TRUE if successful, FALSE if failure.
 */

Boolean
dataPacket::Append(const unsigned char *data, int len)
{
	TRACE(dataPacketModule,1,("In Append\n"));
	if ( _head == _tail && _head == _buf ) { 
		return InsertNewData(data,len);
	}
	else {
		if ( (_buf + MaxPacketSize - _tail) < len ) { 
			// Not enough space.
			TRACE(dataPacketModule,1,("Data larger than MaxPacketSize\n"));
			return FALSE;
		}
		// Copy data and rearrange pointer
		memcpy(_tail,data,len);
		_tail += len;
		len += len;
	}

	TRACE(dataPacketModule,1,("Head: %d Tail: %d Buf: %d\n",_head,_tail,_buf)); 
	return TRUE;
}

/*
 * Function : PreStrip
 *
 * Description : This function deletes a certain length of data
 *	from the front of the buffer, and rearranges the
 *	pointers.
 */

Boolean
dataPacket::PreStrip(int len)
{
	TRACE(dataPacketModule,1,("In PreStrip\n"));
	if ( len > (_tail - _head)) {
		TRACE(dataPacketModule,1,("Stripping more that there is!\n"));
		return FALSE;
	}

	if ( len == 0 )
		return TRUE;

	_head += len; // just rearrange pointer..

	TRACE(dataPacketModule,1,("Head: %d Tail: %d Buf: %d\n",_head,_tail,_buf)); 
	return TRUE;
}

/*
 * Function : PostStrip
 *
 * Description : This function deletes a certain length of data
 *	from the end of the buffer, and rearranges the
 *	pointers.
 */

Boolean
dataPacket::PostStrip(int len)
{
	TRACE(dataPacketModule,1,("In PostStrip\n"));
	if ( len > (_tail - _head)) {
		TRACE(dataPacketModule,1,("Stripping more that there is!\n"));
		return FALSE;
	}

	if ( len == 0 )
		return TRUE;

	_tail -= len; // Just move the tail pointer
	TRACE(dataPacketModule,1,("Head: %d Tail: %d Buf: %d\n",_head,_tail,_buf)); 
	return TRUE;
}

/*
 * Function : Register()
 *
 * Description : This function registers a user with this dataPacket
 *	object by incrementing the _count data member.
 */

void
dataPacket::Register()
{
	_count++;
}

/*
 * Function : Deregister()
 *
 * Description : This function deregisters a user with this dataPacket
 *	object by decrementing the _count data member. If the count
 *	goes to zero, the object is deleted.
 */

void
dataPacket::Deregister()
{
	_count--;
	if (  _count <= 0 )
		delete this;
}

/*
 * Function : InsertNewData
 *
 * Description : Inserts new data into the circular buffer.
 * It assumes and empty buffer. The data is positioned 
 * mid-way in the buffer, to allow for appending and
 * prepending on both sides of the buffer.
 * Returns TRUE if successful, FALSE otherwise.
 */

Boolean	
dataPacket::InsertNewData(const unsigned char *data, int len)
{
	TRACE(dataPacketModule,1,("In InsertNewData\n"));
	if ( len > MaxPacketSize ) {
		TRACE(dataPacketModule,1,("Data larger than MaxPacketSize\n"));
		return FALSE;
	}

	_head = _buf + (( MaxPacketSize - len ) / 2) ;
	memcpy(_head,data,len);
	_tail = _head + len;

	TRACE(dataPacketModule,1,("Head: %d Tail: %d Buf: %d\n",_head,_tail,_buf)); 
	return TRUE;
}

/*
 * Function : Print()
 *
 * Description : An utility function to print the data.
 *	Prints both hex characters, and ascii.
 *	Hex is important for FRMR information.
 */

void
dataPacket::Print()
{
	int len = _tail - _head;
	char msg[100];
	char charmsg[100];

	cout << "\n";
	for ( int i = 0; i < len; i++ ) {
		sprintf(msg+3*i,"%3x ",*(_head+i));
		sprintf(charmsg,"%c",*(_head+i));
	}
	strcpy(msg+3*i,"");
	strcpy(charmsg+i,"\n");	
	cout << msg << "...." << charmsg << flush;
}

