/*
 * File : dllayer.C
 *
 * Description : This file contains the method definitions for
 * the dllayer class.
 */

#include <stream.h>
#include "scheduler.h"
#include "dllayer.h"
#include "phEvents.h"
#include "dlEvents.h"
#include "frame.h"
#include "timerT1.h"
#include "dlDisconnected.h"
#include "types.h"
#include "dataPacket.h"
#include "frame.h"

DEFINE_TRACE_MODULE(dllayerModule,"dllayerModule");

/*
 * Function : dllayer()
 *
 * Description : This is the constructor. Initializes the data
 *	members.
 */

dllayer::dllayer() :
_t1Timer(0),
_rejTimer(0),
_lastFrameRecd(0), _expFBit(0),
_win(MODULO),_remoteBusy(FALSE)
{
}

/*
 * Function : ~dllayer()
 *
 * Description : This is the destructor. Frees memory. 
 * Stops timers.
 */

dllayer::~dllayer()
{
	if ( _sm )
		delete _sm;
	if ( _lastFrameRecd )
		delete _lastFrameRecd;
	StopT1Timer();
	StopRejTimer();
}

/*
 * Function : Init()
 *
 * Description : This function creates a current state machine
 *	object (in the dlDisconnected state). Calls the PH_ActivateReq
 *	primitive on the physical layer.
 */

int
dllayer::Init()
{
	TRACE(dllayerModule,1,("In Init\n"));
	_sm = new dlDisconnected(this);

	// Activate the physical layer
	int errorCode = codeSuccessful;

	_nMinus1->Command(PH_ActivateReq,errorCode);

	if ( errorCode != codeSuccessful )
		cout << "Physical layer activation failed\n";

	return errorCode;
}


/*
 * Function : Command(int cmd, int &errorCode)
 *
 * Description : Handles primitives other than PH_DataInd and
 * DL_DataReq. Invokes ProcessEvent with the primitive as
 * the event.
 */

void 
dllayer::Command(int cmd,int &errorCode)
{
	TRACE(dllayerModule,1,("In Command_int\n"));
	ProcessEvent(cmd,errorCode);
}

/*
 * Function : Command(int cmd, int &errorCode, dataPacket *data)
 *
 * Description : Handles primitives PH_DataInd, DL_DataReq. 
 *	For received dataPacket - it first stores it in _lastFrameRecd
 *	before calling ProcessEvent - so that state machine objects
 *	can access the last frame that was received.
 *	For dataPacket to be transmitted - it first pushes it into
 *	the window's queue, and then calls ProcessEvent.
 */

void 
dllayer::Command(int cmd, int &errorCode, dataPacket *data)
{
	TRACE(dllayerModule,1,("In Command_int_dataPacket*\n"));
	switch (cmd ) 
		{
	case PH_DataInd:
		if ( _lastFrameRecd )
			delete _lastFrameRecd;
		// Create a new frame object with the received
		// dataPacket.
		_lastFrameRecd = new frame(data,FALSE);	
		_lastFrameRecd->PrintParams();

		// For testing - if reception is blocked, just
		// return..

		if ( _blockRecv == FALSE)
			ProcessEvent(PH_DataInd,errorCode);
		break;

	case DL_DataReq:

		// Add new data to be transmitted to window queue.
		// And call ProcessEvent.

		_win.AddFrame(data);
		ProcessEvent(DL_DataReq,errorCode);
		break;
	default:
		TRACE(dllayerModule,1,("Unknown command\n"));
		errorCode = codeGeneralFailure;
		break;
		}
}

/*
 * Function : Command(int cmd, int &errorCode, int reason)
 *
 * Description : Calls base class function.
 */

void 
dllayer::Command(int cmd, int &errorCode, int reason) 
{
	layer::Command(cmd,errorCode,reason);
}

/*
 * Function : ProcessEvent(int event, int &errorCode)
 *
 * Description : This function receives an event and based on
 *	the event type it invokes the appropriate method on
 * 	the current state object.
 *	For the case of a PH_DataInd event - the event is
 *	further analysed for data-link frame types before
 *	invoking the method.
 */

void	
dllayer::ProcessEvent(int event, int &errorCode)
{
	TRACE(dllayerModule,2,("Entering ProcessEvent : %x %d\n",_sm,event));
	_sm->PrintState();
	switch ( event ) 
		{
	case DL_ConnectReq:
		_sm = _sm->DlConnectReq(errorCode);
		break;
	case DL_ConnectRes:
		_sm = _sm->DlConnectRes(errorCode);
		break;
	case DL_ResetReq:
		_sm = _sm->DlResetReq(errorCode);
		break;
	case DL_ResetRes:
		_sm = _sm->DlResetRes(errorCode);
		break;
	case DL_DataReq:
		_sm = _sm->DlDataReq(errorCode);
		break;
	case DL_DisconnectReq:
		_sm = _sm->DlDisconnectReq(errorCode);
		break;
	case PH_DeactivateInd:
		_sm = _sm->PhDeactivateInd(errorCode);
		break;
	case PH_DataInd:
		// Check for unsolicited f-bit

		if ( (_lastFrameRecd->IsCmd() == FALSE) &&
				(_lastFrameRecd->GetPf() == 1) ) { 
			if ( _expFBit == 0 ) {
				// We got unsolicited F-BIT
				_sm = _sm->HandleUnsolFBit(errorCode);
				break;
			}
			else 
				// We got a solicited F bit, now clear it.
				_expFBit = 0;
		}
		switch ( _lastFrameRecd->GetFrameType() )		
			{
		case INVALID:
			errorCode = codeSuccessful;
			break;
		case UNDEFINED:
			_sm = _sm->HandleUndefined(errorCode);
			break;
		case SABM:
			_sm = _sm->HandleSabm(errorCode); 
			break;
		case UA:
			_sm = _sm->HandleUa(errorCode);
			break;
		case DISC:
			_sm = _sm->HandleDisc(errorCode);
			break;
		case DM:
			_sm = _sm->HandleDm(errorCode);
			break;
		case FRMR:
			_sm = _sm->HandleFrmr(errorCode);
			break;
		case IFRAME:
			_sm = _sm->HandleIframe(errorCode);
			break;
		case RNR:
			_sm = _sm->HandleRnr(errorCode);
			break;
		case REJ:
			_sm = _sm->HandleRej(errorCode);
			break;
		case RR:	
			_sm = _sm->HandleRr(errorCode);
			break;
		case IfrLengthExcess:
			_sm = _sm->HandleIfrLengthExcess(errorCode);
			break;
		case IncorrectLength:
			_sm = _sm->HandleIncorrectLength(errorCode);
			break;
		default:
			_sm = _sm->HandleUndefined(errorCode);
			TRACE(dllayerModule,2,("Unknown type in frame\n"));
			break;
			}
		break;
	case t1TimerExp:
		_sm = _sm->HandleT1Timer(errorCode);
		break;
	case rejTimerExp:
		_sm = _sm->HandleRejTimer(errorCode);
		break;
	default:
		errorCode = codeGeneralFailure;
		TRACE(dllayerModule,2,("Unknown Event\n"));
		break;
		}
	_sm->PrintState();
	TRACE(dllayerModule,2,("Exiting ProcessEvent : %x\n",_sm));
}

/*
 * Function : StartT1Timer()
 *
 * Description : This functions starts the _t1Timer. If 'continueRetry'
 *	is TRUE - then the timer is continuing from previous retry attemps.
 *	Else, it is a fresh timer - with retry count equal to zero.
 */

void 
dllayer::StartT1Timer(Boolean continueRetry)
{
	scheduler *sch = scheduler::GetScheduler();

	TRACE(dllayerModule,1,("In StartT1Timer Restart : %d\n",continueRetry));
	if ( continueRetry == TRUE ) {
		// timer's already running,override it

		if ( _t1Timer == 0 ) { // Error!
			TRACE(dllayerModule,1,("Timer T1 is not running\n"));
			return;
		}
		_t1Timer = StartTimer(t1TimerExp);
		_t1RetryCount++;
	}
	else {
		if ( _t1Timer != 0 ) {
			TRACE(dllayerModule,1,("Timer is running!\n"));
			StopT1Timer();
		}
		_t1Timer = StartTimer(t1TimerExp);
		_t1RetryCount = 0;
	}
}


/*
 * Function : StartRejTimer()
 *
 * Description : This functions starts the _rejTimer. If 'continueRetry'
 *	is TRUE - then the timer is continuing from previous retry attemps.
 *	Else, it is a fresh timer - with retry count equal to zero.
 */

void 
dllayer::StartRejTimer(Boolean continueRetry)
{
	scheduler *sch = scheduler::GetScheduler();

	TRACE(dllayerModule,1,("In StartRejTimer Restart : %d\n",continueRetry));
	if ( continueRetry == TRUE ) {
		// timer's already running,override it

		if ( _rejTimer == 0 ) { // Error!
			TRACE(dllayerModule,1,("Timer rejTimer is not running\n"));
			return;
		}
		_rejTimer = StartTimer(rejTimerExp);
		_rejRetryCount++;
	}
	else {
		if ( _rejTimer != 0 ) {
			TRACE(dllayerModule,1,("Timer is running!\n"));
			StopRejTimer();
		}
		_rejTimer = StartTimer(rejTimerExp);
		_rejRetryCount = 0;
	}
}


/*
 * Function : StartTimer()
 *
 * Description : This function creates a new timerT1 object and
 *	returns a pointer to it. It also adds the timer to be 
 * 	scheduled.
 */

timerT1 *
dllayer::StartTimer(int event) 
{
	timerT1 *t = new timerT1(event,this);
	scheduler *sch = scheduler::GetScheduler();
	sch->AddEventHandler(t);
	return t;
}

/*
 * Function : StopT1Timer()
 *
 * Description : This function stops the _t1Timer if it is running,
 *	removes it from the scheduler - and deletes it.
 */

void
dllayer::StopT1Timer()
{
	if ( _t1Timer ) {
		_t1Timer->Disable();
		_t1Timer = 0;
		_t1RetryCount = 0;
	}
}


/*
 * Function : StopRejTimer()
 *
 * Description : This function stops the _rejTimer if it is running,
 *	removes it from the scheduler - and deletes it.
 */

void
dllayer::StopRejTimer()
{
	if ( _rejTimer ) {
		_rejTimer->Disable();
		_rejTimer = 0;
		_rejRetryCount = 0;
	}
}
/*
 * Function : ResetStateVariables()
 *
 * Description : Stops timers, and resets the WINDOW
 */

void 
dllayer::ResetStateVariables()
{
	// Stop any timers
	_expFBit = 0;
	StopT1Timer();
	StopRejTimer();
	_win.Reset(); // Drops any i-frames in the queue.
}
/*
 * Function : XmitFrmr()
 *
 * Description : This method transmits a FRMR frame. It
 * adds the info. in _frmrInfo[] as the information
 * field in the FRMR frame. The _frmrInfo array would
 * have been set elsewhere prior to calling XmitFrmr().
 */

void		
dllayer::XmitFrmr(int pf, int &errorCode)
{
	TRACE(dllayerModule,1,("Transmitting FRMR %c %c %c\n", 
		_frmrInfo[0],_frmrInfo[1],_frmrInfo[2]));

	dataPacket *frmrData = new dataPacket(_frmrInfo,3);
	frame fr(frmrData,TRUE); // Instantiate a frame
	fr.SetParams(FRMR,pf); // Set to FRMR frame type
	fr.FormFrame(FALSE); // Form the response frame
	_nMinus1->Command(PH_DataReq, errorCode, fr.GetDataPacket());
}

/*
 * Function : XmitFrame()
 *
 * Description : Utility function to transmit a FRAME.
 * Sets frame parameters as provided, and invokes
 * PH_DataReq on the physical layer.
 */
 
void
dllayer::XmitFrame(int &errorCode, Boolean cmd, int frameType, 
int pf, int nr, int ns)  
{
	frame fr;
	fr.SetParams(frameType,pf,nr,ns);
	fr.FormFrame(cmd);
	_nMinus1->Command(PH_DataReq, errorCode, fr.GetDataPacket()); 
	if ( errorCode == codeSuccessful ) 
		if ( cmd == TRUE )
			_expFBit = pf;
	return;
}

